// -----------------------------------------------
// Digger for .NET by Lutz Roeder, June 2000.
// Graphics and Levels by Alexander Lang and Jan Fricke.
// roeder@aisto.com
// http://www.aisto.com/roeder
// -----------------------------------------------
namespace Digger
{
	using System;
	using System.IO;
	using System.Collections;
	using System.Drawing;
	using System.Drawing.Imaging;
	using System.Windows.Forms;

	internal class ApplicationWindow : Form
	{
		private GameWindow window = new GameWindow();

		public ApplicationWindow() 
		{
			this.Icon = new Icon(GetType().Assembly.GetManifestResourceStream("Application.ico"));
			this.Text = ".NET Digger";
			this.FormBorderStyle = FormBorderStyle.FixedDialog;
			this.ClientSize = this.window.ClientSize;
			this.MaximizeBox = false;    
			this.Controls.Add(this.window);
		}

		protected override void OnGotFocus(EventArgs e)
		{
			base.OnGotFocus(e);
			this.window.Focus();
		}

		public static void Main(string[] args)
		{
			Application.Run(new ApplicationWindow());
		}
	}

	public class GameWindow : Panel
	{ 
		private int magnification = 1;
		private Bitmap[] sprites = new Bitmap[34];
		private Font font;
		private Timer timer;
		private Level level = new Level();
		private int tick = 0;
		private int lives = 20;
		private bool[] keysRelease = new bool[4];
		private bool[] keysDelay = new bool[4];  
  
		public GameWindow()
		{
			this.ClientSize = MagnifyRectangle(0, 0, 320, 256).Size;
			this.BorderStyle = BorderStyle.FixedSingle;
			this.font = new Font(MagnifyBitmap(new Bitmap(GetType().Assembly.GetManifestResourceStream("Font.png"))), MagnifyRectangle(0, 0, 8, 8).Size);

			Bitmap sprites = MagnifyBitmap(new Bitmap(GetType().Assembly.GetManifestResourceStream("Sprite.png")));
			Rectangle cursor = MagnifyRectangle(0, 0, 16, 16);
			for (int i = 0; i < 34; i++)
			{
				this.sprites[i] = sprites.Clone(cursor, PixelFormat.Format32bppArgb);
				cursor.X += cursor.Width;
			}

			this.level.Load();
			this.timer = new Timer();
			this.timer.Interval = 40;
			this.timer.Tick += new EventHandler(this.Timer_Tick);
			this.timer.Start();
		}

		protected override void OnKeyDown(KeyEventArgs e)
		{
			if (e.KeyCode == Keys.Left) { level.Digger.Keys[0] = true; keysDelay[0] = false; }
			if (e.KeyCode == Keys.Right) { level.Digger.Keys[1] = true; keysDelay[1] = false; }    
			if (e.KeyCode == Keys.Up) { level.Digger.Keys[2] = true; keysDelay[2] = false; }
			if (e.KeyCode == Keys.Down) { level.Digger.Keys[3] = true; keysDelay[3] = false; }        
			if ((e.KeyCode == Keys.Escape) && (lives > 0)) { lives--; level.Load(); }
			if ((e.KeyCode == Keys.Home) && (level.Number < 29)) { level.Number++; level.Load(); }
		}

		protected override void OnKeyUp(KeyEventArgs e)
		{
			if (e.KeyCode == Keys.Left) 
				if (keysDelay[0]) level.Digger.Keys[0] = false; else keysRelease[0] = true;
			if (e.KeyCode == Keys.Right) 
				if (keysDelay[1]) level.Digger.Keys[1] = false; else keysRelease[1] = true;
			if (e.KeyCode == Keys.Up) 
				if (keysDelay[2]) level.Digger.Keys[2] = false; else keysRelease[2] = true;
			if (e.KeyCode == Keys.Down) 
				if (keysDelay[3]) level.Digger.Keys[3] = false; else keysRelease[3] = true;
		}

		private void Timer_Tick(object sender, EventArgs pe)
		{
			this.tick++;
			this.level.Tick1();

			if ((this.tick % 2) == 0)
			{
				if (this.keysRelease[0]) { this.level.Digger.Keys[0] = this.keysRelease[0] = false; }
				if (this.keysRelease[1]) { this.level.Digger.Keys[1] = this.keysRelease[1] = false; }
				if (this.keysRelease[2]) { this.level.Digger.Keys[2] = this.keysRelease[2] = false; }
				if (this.keysRelease[3]) { this.level.Digger.Keys[3] = this.keysRelease[3] = false; }                  
				this.level.Tick4();
			}

			this.Invalidate(this.MagnifyRectangle(0, 8, 320, 16));

			for (int y = 0; y < 14; y++)
			{
				for (int x = 0; x < 20; x++)
				{
					Sprite sprite = level.GetSprite(x, y);
					if (sprite.Invalidate)
					{
						Rectangle rectangle = MagnifyRectangle(x * 16, y * 16 + 32, 16, 16);
						this.Invalidate(rectangle);
						sprite.Invalidate = false;
					}
				}
			}

			this.Update();
		}

		protected override void OnPaintBackground(PaintEventArgs e) { }

		protected override void OnPaint(PaintEventArgs e)
		{
			if ((e.ClipRectangle.Width != 0) && (e.ClipRectangle.Height != 0))
			{
				Graphics graphics = e.Graphics;

				SolidBrush blueBrush = new SolidBrush(Color.FromArgb(0x04, 0x02, 0x8F));
				graphics.FillRectangle(blueBrush, MagnifyRectangle(0, 0, 320, 8));
				graphics.FillRectangle(blueBrush, MagnifyRectangle(0, 24, 320, 8));
				SolidBrush cyanBrush = new SolidBrush(Color.FromArgb(0x2D, 0xE7, 0xC0));
				graphics.FillRectangle(cyanBrush, MagnifyRectangle(0, 2, 320, 4));
				graphics.FillRectangle(cyanBrush, MagnifyRectangle(0, 26, 320, 4));
	
				font.Reset();
				font.Newline();
				font.Write(graphics, "  CAVE:  " + (level.Number + 1).ToString("D2") + " TIME:  " + level.Time.ToString("D5") + " DIAMONDS:  " + level.Diamonds.ToString("D2") + "  ");
				font.Newline();
				font.Write(graphics, "  LIVES: " + lives.ToString("D2") + " SCORE: " + level.Score.ToString("D5") + " COLLECTED: " + level.Collected.ToString("D2") + "  ");
				font.Newline();
	
				for (int y = 0; y < 14; y++)
				{
					for (int x = 0; x < 20; x++)
					{
						Rectangle rectangle = this.MagnifyRectangle(x * 16, y * 16 + 32, 16, 16);
						if (e.ClipRectangle.IntersectsWith(rectangle)) 
							graphics.DrawImageUnscaled(sprites[level.GetSprite(x, y).ImageIndex], rectangle.X, rectangle.Y);
					}
				}
			}
		}

		private Rectangle MagnifyRectangle(int x, int y, int dx, int dy)
		{
			return new Rectangle(x * magnification, y * magnification, dx * magnification, dy * magnification);
		}

		private Bitmap MagnifyBitmap(Bitmap source)
		{
			Bitmap target = new Bitmap(source.Width * magnification, source.Height * magnification);
			for (int x = 0; x < source.Width; x++)
				for (int y = 0; y < source.Height; y++)
					for (int ix = 0; ix < magnification; ix++)
						for (int iy = 0; iy < magnification; iy++)
							target.SetPixel(x * magnification + ix, y * magnification + iy, source.GetPixel(x, y));
			return target;
		}

		private enum Directions { Left, Right, Up, Down, None }
  
		private class Position
		{
			public int x, y;
  
			public Position() { x = 0; y = 0; }
			public Position(int x, int y) { this.x = x; this.y = y; }
			public Position(Position p) { this.x = p.x; this.y = p.y; }
			public bool Equal(Position p) { return ((this.x == p.x) && (this.y == p.y)); }
		}
  
		private class Sprite 
		{
			public bool Invalidate = true;
			public virtual int ImageIndex { get { return 0; } }
		}
  
		private class Ground : Sprite
		{ 
			public override int ImageIndex { get { return 2; } }
		}
  
		private class Ghost : Sprite 
		{
			public Directions Direction = Directions.None;
			public bool IsAlive = true;
  
			public override int ImageIndex
			{
				get
				{
					if (this.Direction == Directions.Left) return 4;
					if (this.Direction == Directions.Right) return 5;
					if (this.Direction == Directions.Up) return 6;
					if (this.Direction == Directions.Down) return 3;
					return 3;
				}
			}
  
			public virtual void Move(Level level) { }
  
			public void Die()
			{
				this.IsAlive = false;
			}
  
			public void Blast(Level level)
			{
				if (this.IsAlive)
				{
	  				Position p = level.GetSpritePosition(this);
	  
					for (int y = p.y - 1; y <= p.y + 1; y++)
						for (int x = p.x - 1; x <= p.x + 1; x++)
						{
							if ((x > 0) && (x < 19) && (y > 0) && (y < 13))
							{
								if (level.GetSprite(x, y).GetType() == typeof(Digger))
								{
									Digger d = (Digger) level.GetSprite(x, y);
									level.Digger.Die();
								}
								else
								{
									if (level.GetSprite(x, y).GetType().BaseType == typeof(Ghost))
									{
										Ghost ghost = (Ghost) level.GetSprite(x, y);
										ghost.Die();
										level.Score += 99;
									}
	  
									level.SetSprite(x, y, new Nothing());
								}
							}
						}
	  
					this.Die();
				}
			}
		}
  
		private class Ghost180 : Ghost
		{
			public override void Move(Level level)
			{
				if (this.IsAlive)
				{
	  				Position p = level.GetSpritePosition(this);
					Position[] w = { new Position(p), new Position(p), new Position(p), new Position(p) };
	  
					if (this.Direction == Directions.Left)  { w[0].x--; w[1].x++; }
					if (this.Direction == Directions.Right) { w[0].x++; w[1].x--; }
					if (this.Direction == Directions.Up)    { w[0].y--; w[1].y++; }
					if (this.Direction == Directions.Down)  { w[0].y++; w[1].y--; }
	  
					for (int i = 0; i < 4; i++)
					{
						if (!p.Equal(w[i]))
						{
							Position d = new Position(w[i]);
	        
							// Digger
							if (level.GetSprite(d.x, d.y).GetType() == typeof(Digger)) 
								level.Digger.Die();
	  
							// Nothing
							if (level.GetSprite(d.x, d.y).GetType() == typeof(Nothing))
							{
								if (d.x < p.x) Direction = Directions.Left;
								if (d.x > p.x) Direction = Directions.Right;
								if (d.y < p.y) Direction = Directions.Up;
								if (d.y > p.y) Direction = Directions.Down;
	          
								level.SetSprite(d.x, d.y, this);
								level.SetSprite(p.x, p.y, new Nothing());
	  
								return;
							}
						}
					}
				}
			}
		}
  
		private class Ghost90L : Ghost
		{
			public override void Move(Level level)
			{
				if (this.IsAlive)
				{
	  
					Position p = level.GetSpritePosition(this);
					Position[] w = { new Position(p), new Position(p), new Position(p), new Position(p) };
	  
					if (this.Direction == Directions.Left)  { w[0].x--; w[1].y++; w[2].y--; w[3].x++; }
					if (this.Direction == Directions.Right) { w[0].x++; w[1].y--; w[2].y++; w[3].x--; }
					if (this.Direction == Directions.Up)    { w[0].y--; w[1].x--; w[2].x++; w[3].y++; }
					if (this.Direction == Directions.Down)  { w[0].y++; w[1].x++; w[2].x--; w[3].y--; }
	  
					for (int i = 0; i < 4; i++)
					{
						if (!p.Equal(w[i]))
						{
							Position d = new Position(w[i]);
	        
							// Digger
							if (level.GetSprite(d.x, d.y).GetType() == typeof(Digger)) level.Digger.Die();
	  
							// Nothing
							if (level.GetSprite(d.x, d.y).GetType() == typeof(Nothing))
							{
								if (d.x < p.x) Direction = Directions.Left;
								if (d.x > p.x) Direction = Directions.Right;
								if (d.y < p.y) Direction = Directions.Up;
								if (d.y > p.y) Direction = Directions.Down;
								level.SetSprite(d.x, d.y, this);
								level.SetSprite(p.x, p.y, new Nothing());
								return;
							}
						}
					}
				}
			}
		}
  
		private class Ghost90LR : Ghost
		{
			public Directions Lastturn = Directions.None;
  
			public override void Move(Level level)
			{
  				if (this.IsAlive)
  				{
					Position p = level.GetSpritePosition(this);
					Position[] w = { new Position(p), new Position(p), new Position(p), new Position(p) };
	  
					if (this.Direction == Directions.Left)
					{
						w[0].x--; w[3].x++;
						if (Lastturn == Directions.Left) { w[1].y--; w[2].y++; } 
						else { w[1].y++; w[2].y--; };
					}
	  
					if (this.Direction == Directions.Right)
					{
						w[0].x++; w[3].x--;
						if (Lastturn == Directions.Left) { w[1].y++; w[2].y--; } 
						else { w[1].y--; w[2].y++; };
					}
	  
					if (this.Direction == Directions.Up)
					{
						w[0].y--; w[3].y++;
						if (Lastturn == Directions.Left) { w[1].x++; w[2].x--; } 
						else { w[1].x--; w[2].x++; };
					}
	  
					if (this.Direction == Directions.Down)
					{
						w[0].y++; w[3].y--;
						if (Lastturn == Directions.Left) { w[1].x--; w[2].x++; } 
						else { w[1].x++; w[2].x--; };
					}
	     
					for (int i = 0; i < 4; i++)
					{
						if (!p.Equal(w[i]))
						{
							Position d = new Position(w[i]);
	        
							// Digger
							if (level.GetSprite(d.x, d.y).GetType() == typeof(Digger)) level.Digger.Die();
	  
							// Nothing
							if (level.GetSprite(d.x, d.y).GetType() == typeof(Nothing))
							{
								Directions LastDirection = Direction;
	  
								if (d.x < p.x) Direction = Directions.Left;
								if (d.x > p.x) Direction = Directions.Right;
								if (d.y < p.y) Direction = Directions.Up;
								if (d.y > p.y) Direction = Directions.Down;
	  
								switch (LastDirection)
								{
									case Directions.Left:
										if (this.Direction == Directions.Down)  Lastturn = Directions.Left; 
										if (this.Direction == Directions.Up)    Lastturn = Directions.Right;
										break;
									case Directions.Right:
										if (this.Direction == Directions.Down)  Lastturn = Directions.Right; 
										if (this.Direction == Directions.Up)    Lastturn = Directions.Left;
										break;
									case Directions.Up:
										if (this.Direction == Directions.Left)  Lastturn = Directions.Left; 
										if (this.Direction == Directions.Right) Lastturn = Directions.Right;
										break;
									case Directions.Down:
										if (this.Direction == Directions.Left)  Lastturn = Directions.Right; 
										if (this.Direction == Directions.Right) Lastturn = Directions.Left;
										break;
								}        
	  
								level.SetSprite(d.x, d.y, this);
								level.SetSprite(p.x, p.y, new Nothing());
	  
								return;
							}
						}
					}
				}
			}
		}
  
		private class Ghost90R : Ghost
		{
			public override void Move(Level level)
			{
				if (this.IsAlive)
				{
					Position p = level.GetSpritePosition(this);
					Position[] w = { new Position(p), new Position(p), new Position(p), new Position(p) };
	  
					if (this.Direction == Directions.Left)  { w[0].x--; w[1].y--; w[2].y++; w[3].x++; }
					if (this.Direction == Directions.Right) { w[0].x++; w[1].y++; w[2].y--; w[3].x--; }
					if (this.Direction == Directions.Up)    { w[0].y--; w[1].x++; w[2].x--; w[3].y++; }
					if (this.Direction == Directions.Down)  { w[0].y++; w[1].x--; w[2].x++; w[3].y--; }
	  
					for (int i = 0; i < 4; i++)
					{
						if (!p.Equal(w[i]))
						{
							Position d = new Position(w[i]);
	        
							// Digger
							if (level.GetSprite(d.x, d.y).GetType() == typeof(Digger))
								level.Digger.Die();
	  
							// Nothing
							if (level.GetSprite(d.x, d.y).GetType() == typeof(Nothing))
							{
								if (d.x < p.x) Direction = Directions.Left;
								if (d.x > p.x) Direction = Directions.Right;
								if (d.y < p.y) Direction = Directions.Up;
								if (d.y > p.y) Direction = Directions.Down;
	          
								level.SetSprite(d.x, d.y, this);
								level.SetSprite(p.x, p.y, new Nothing());
	  
								return;
							}
						}
					}
				}
			}
		}
  
		private class Wall : Sprite 
		{
			public override int ImageIndex { get { return 14; } }
		}
  
		private class Digger : Sprite 
		{
			public bool[] Keys = new bool[4];
			private Directions direction = Directions.None;
			private int step = 0;
			private bool isAlive = true;
			private bool[] stone = new bool[2];
  
			public override int ImageIndex
			{
				get
				{
					if (this.isAlive)
					{
						int[] left = { 16, 17, 18, 19, 18, 17 };
						int[] right = { 20, 21, 22, 23, 22, 21 };
						int[] up = { 24, 25 };
						int[] down = { 26, 27 };
						int[] center = { 15, 15, 15, 15, 15, 15, 15, 15, 28, 28, 15, 15, 28, 28, 15, 15, 15, 15, 15, 15, 29, 29, 30, 30, 29, 29, 15, 15, 15, 15 };
						if ((this.direction == Directions.Left) && (step <= 5)) return left[step];
						if ((this.direction == Directions.Right) && (step <= 5)) return right[step];
						if ((this.direction == Directions.Up) && (step <= 1)) return up[step];
						if ((this.direction == Directions.Down) && (step <= 1)) return down[step];
						return center[step];
					}
					
					return 31;
				}
			}
  
			public void Animate(Level level, int x, int y)
			{
				if (this.isAlive)
				{
	  				switch (this.direction)
					{
						case Directions.Left:
							step++;
							if (step >= 6) step = 0;
							level.Digger.Invalidate = true;
							break;
	  
						case Directions.Right:
							step++;
							if (step >= 6) step = 0;
							level.Digger.Invalidate = true;
							break;
	  
						case Directions.Up:
							step++;
							if (step >= 2) step = 0;
							level.Digger.Invalidate = true;
							break;
	  
						case Directions.Down:
							step++;
							if (step >= 2) step = 0;
							level.Digger.Invalidate = true;
							break;
	  
						case Directions.None:
							step++;
							if (step >= 30) step = 0;
							level.Digger.Invalidate = true;
							break;
					}
				}
			}
  
			public void Move(Level level)
			{
				if (this.isAlive)
				{
					Position p = level.GetSpritePosition(this);
					Position d = new Position(p);
					Position z = new Position(d);
	  
					direction = Directions.None;
					if (Keys[0])
					{
						z.x--; direction = Directions.Left;
					} 
					else
					{
						stone[0] = false;
						if (Keys[1])
						{
							z.x++; direction = Directions.Right;
						}
						else
						{
							stone[1] = false;
							if (Keys[2])
							{
								z.y--; direction = Directions.Up;
							}
							else 
							{
								if (Keys[3])
								{ 
									z.y++; direction = Directions.Down;
								}
							}
						} 
					}
	  
					if (!d.Equal(z))
					{
						// Nothing
						if (level.GetSprite(z.x, z.y).GetType() == typeof(Nothing))
							level.SetSprite(d.x, d.y, this);
	  
						// Diamond
						if (level.GetSprite(z.x, z.y).GetType() == typeof(Diamond))
						{
							level.Collected += 1;
							level.Score += 3;
						}
	  
						// Stone
						if (level.GetSprite(z.x, z.y).GetType() == typeof(Stone))
						{
							if ((z.x > d.x) && (level.GetSprite(z.x + 1, z.y).GetType() == typeof(Nothing)))
							{
								if (stone[1])
								{
									level.SetSprite(d.x + 2, d.y, level.GetSprite(d.x + 1, d.y));
									level.SetSprite(d.x + 1, d.y, new Nothing());
								}
								stone[1] = !stone[1];
							}
	  
							if ((z.x < d.x) && (level.GetSprite(z.x - 1, z.y).GetType() == typeof(Nothing)))
							{
								if (stone[0])
								{
									level.SetSprite(d.x - 2, d.y, level.GetSprite(d.x - 1, d.y));
									level.SetSprite(d.x - 1, d.y, new Nothing());
								}
								stone[0] = !stone[0];
							}
						}
	  
						// Ground
						if (level.GetSprite(z.x, z.y).GetType() == typeof(Nothing) ||
							level.GetSprite(z.x, z.y).GetType() == typeof(Ground) ||
							level.GetSprite(z.x, z.y).GetType() == typeof(Diamond))
						{
							level.SetSprite(z.x, z.y, this);
							level.SetSprite(d.x, d.y, new Buffer());
						}
	  
						// Exit
						if (level.GetSprite(z.x, z.y).GetType() == typeof(Exit)) level.Exit();
	  
						// Ghost
						if (level.GetSprite(z.x, z.y).GetType().BaseType == typeof(Ghost)) level.Digger.Die();
					}
	  
					this.Animate(level, z.x, z.y);
				}
			}
  
			public void Die()
			{
				this.isAlive = false;
				this.Invalidate = true;
			}
		}
  
		private class Exit : Sprite 
		{
			public override int ImageIndex { get { return 32; } }
		}
  
		private class Changer : Sprite
		{
			public override int ImageIndex{ get { return 33; } }
		}
  
		private class Gravitation : Sprite
		{
			public void Simulate(Level level, int x, int y)
			{
				int dx = x;
				int dy = y;
  
				// Nothing
				if (level.GetSprite(x, y + 1).GetType() == typeof(Nothing))
				{
					dy = y + 1;
				}
				else
				{
					// Stone or diamond
					if ((level.GetSprite(x, y + 1).GetType() == typeof(Stone)) || (level.GetSprite(x, y + 1).GetType() == typeof(Diamond)))
					{
						if (level.GetSprite(x - 1, y + 1).GetType() == typeof(Nothing) && level.GetSprite(x - 1, y).GetType() == typeof(Nothing))
						{
							dx = x - 1;
							dy = y + 1;
						}
						else
						{
							if (level.GetSprite(x + 1, y + 1).GetType() == typeof(Nothing) && level.GetSprite(x + 1, y).GetType() == typeof(Nothing))
							{
								dx = x + 1;
								dy = y + 1;
							}
						}
					}
  
					// Changer
					if ((level.GetSprite(x, y + 1).GetType() == typeof(Changer)) && (level.GetSprite(x, y).GetType() == typeof(Stone)) && (level.GetSprite(x, y + 2).GetType() == typeof(Nothing)))
						dy = y + 2;
				}
  
				// Mark
				if ((dx != x) || (dy != y)) level.SetSprite(dx, dy, new Marker());
			}
  
			public void Move(Level level, int x, int y)
			{
				int dx = x;
				int dy = y;
  
				// Follow gravitation
				if (level.GetSprite(x, y + 1).GetType() == typeof(Marker))
				{
					dy = y + 1;
				}
				else
				{
					// Falls to the left or to the right
					if (level.GetSprite(x, y + 1).GetType() == typeof(Stone) || level.GetSprite(x, y + 1).GetType() == typeof(Diamond) || level.GetSprite(x, y + 1).GetType() == typeof(Nothing))
					{
						if (level.GetSprite(x - 1, y + 1).GetType() == typeof(Marker) && (level.GetSprite(x - 1, y).GetType() == typeof(Nothing) || level.GetSprite(x - 1, y).GetType() == typeof(Marker)))
						{
							dx = x - 1;
							dy = y + 1;
						}
						else
						{
							if (level.GetSprite(x + 1, y + 1).GetType() == typeof(Marker) && (level.GetSprite(x + 1, y).GetType() == typeof(Nothing) || level.GetSprite(x + 1, y).GetType() == typeof(Marker)))
							{
								dx = x + 1;
								dy = y + 1;
							}
						}
					}
  
					// Changer
					if (level.GetSprite(x, y + 1).GetType() == typeof(Changer) && level.GetSprite(x, y).GetType() == typeof(Stone) && level.GetSprite(x, y + 2).GetType() == typeof(Marker)) dy = y + 2;
				}
  
				// Move
				if ((dx != x) || (dy != y))
				{
					if ((dy - y) == 2)
					{
						// changer
						level.SetSprite(dx, dy, new Diamond());
					}
					else
					{
						level.SetSprite(dx, dy, level.GetSprite(x, y));
						if (level.GetSprite(dx, dy).GetType() == typeof(UvStone))
							level.SetSprite(dx, dy, new Stone());
					}
  
					level.SetSprite(x, y, new Nothing());      
  
					// Kill digger if necessary
					if (level.GetSprite(dx, dy + 1).GetType() == typeof(Digger)) level.Digger.Die();
  
					if (level.GetSprite(dx, dy + 1).GetType().BaseType == typeof(Ghost))
					{
						Ghost ghost = (Ghost) level.GetSprite(dx, dy + 1);
						ghost.Blast(level);
					}
				}
			}
		}
  
		private class Marker : Gravitation
		{
			public override int ImageIndex { get { return 31; } }
		}
  
		private class Stone : Gravitation
		{
			public override int ImageIndex { get { return 1; } }
		}
  
		private class Diamond : Gravitation
		{
			private static int blinkCounter = 0;
  
			public static void Blink()
			{
				blinkCounter++;
				blinkCounter = (blinkCounter == 6) ? 0 : blinkCounter;
			}
  
			public override int ImageIndex
			{
				get
				{
					return (13 - ((blinkCounter + 4 * 1) % 6));
				}
			}
		}
  
		private class Nothing : Sprite { }
		private class Buffer : Sprite { }
		private class UvStone : Gravitation { }
  
		private class Level
		{
			private Byte[] data;
			private Sprite[,] map = new Sprite[20, 14];
			private Ghost[] ghosts = new Ghost[16];
			public Digger Digger;
			public int Number;
			public int Diamonds;
			public int Collected;
			public int Time;
			public int Score = 0;
  
			public Level()
			{
				Stream stream = GetType().Assembly.GetManifestResourceStream("Level.bin");
				data = new Byte[stream.Length];
				stream.Read(data, 0, (int) stream.Length);
  
				Number = 0;
			}
  
			public void Load()
			{
				Byte[] level = new Byte[156];
				for (int i = 0; i < 156; i++)
					level[i] = data[Number * 156 + i];
      
				for (int y = 0; y < 14; y++)
					for (int x = 0; x < 10; x++)
					{
						CreateSprite(x * 2, y , (byte) (level[y * 10 + x] >> 4));
						CreateSprite(x * 2 + 1, y , (byte) (level[y * 10 + x] & 0x0F));
					}
  
				Digger = (Digger) map[level[145], level[146]-2];
  
				for (int g = 0; g < 16; g++) ghosts[g] = null;
				int Ghost = 0;
				for (int y = 0; y < 14; y++)
					for (int x = 0; x < 20; x++)
					{
						if (GetSprite(x, y).GetType().BaseType == typeof(Ghost))
						{
							Ghost g = (Ghost) GetSprite(x, y);
							ghosts[Ghost] = g;
							int GhostInfo = level[0x94 + (Ghost >> 1)];
							if ((Ghost & 1) != 0)
							{
								GhostInfo = GhostInfo & 0x0F;
								if (g.GetType() == typeof(Ghost90LR)) ((Ghost90LR) g).Lastturn = Directions.Right;
							}
							else
							{
								GhostInfo = GhostInfo >> 4;
								if (g.GetType() == typeof(Ghost90LR)) ((Ghost90LR) g).Lastturn = Directions.Left;
							}
							Directions[] d = { Directions.Down, Directions.Up, Directions.Right, Directions.Left };
							g.Direction = d[GhostInfo];
							Ghost++;
						}
					}
  
				Diamonds = (level[147] / 0x10) * 10 + (level[147] % 0x10);
				Collected = 0;
				Time = 5000;
			}
  
			public void Exit()
			{ 
				if (Collected < Diamonds) return;
				if (Number >= 29) return;
				Number++;
				Load();
			}
  
			public void CreateSprite(int x, int y, int id)
			{
				Type[] t = { typeof(Nothing), typeof(Stone), typeof(Ground), typeof(Ghost180), null, typeof(Diamond), typeof(Wall), typeof(Ghost90L), null, typeof(UvStone), typeof(Digger), typeof(Ghost90LR), typeof(Exit), null, typeof(Changer), typeof(Ghost90R) };
				SetSprite(x, y, (Sprite) Activator.CreateInstance(t[id]));
			} 
  
			public void SetSprite(int x, int y, Sprite s)
			{
				map[x, y] = s;
				s.Invalidate = true;
			}
  
			public Sprite GetSprite(int x, int y)
			{
				return map[x, y];
			}
  
			public Position GetSpritePosition(Sprite s)
			{
				for (int x = 0; x < 20; x++)
					for (int y = 0; y < 14; y++)
						if (GetSprite(x, y) == s)
							return new Position(x, y);
				return new Position();
			}
  
			public void Tick1()
			{
				Diamond.Blink();
				for (int y = 0; y < 14; y++)
					for (int x = 0; x < 20; x++)
						if (GetSprite(x, y).GetType() == typeof(Diamond))
							GetSprite(x, y).Invalidate = true;
			}
  
			public void Tick4()
			{
				// Turn buffers into nothing
				for (int y = 13; y >= 0; y--)
					for (int x = 19; x >= 0; x--)
						if (this.map[x, y].GetType() == typeof(Buffer))
							this.SetSprite(x, y, new Nothing());
  
				// Digger
				this.Digger.Move(this);
  
				// Stones and Diamonds (UVStones?)
				for (int y = 13; y >= 0; y--)
					for (int x = 19; x >= 0; x--)
						if (map[x, y].GetType() == typeof(Stone) || map[x, y].GetType() == typeof(Diamond) || map[x, y].GetType() == typeof(UvStone))
						{
							Gravitation gravitation = (Gravitation) map[x, y];
							gravitation.Simulate(this, x, y);
						}
  
				for (int y = 13; y >= 0; y--)
					for (int x = 19; x >= 0; x--)
						if (map[x, y].GetType() == typeof(Stone) || map[x, y].GetType() == typeof(Diamond) || map[x, y].GetType() == typeof(UvStone))
						{
							Gravitation gravitation = (Gravitation) map[x, y];
							gravitation.Move(this, x, y);
						}
  
				for (int g = 0; g < 16; g++)
					if (ghosts[g] != null) ghosts[g].Move(this);
  
				if (this.Time > 0) this.Time--;
				if (this.Time == 0) this.Digger.Die();
			}
		}
	}

	internal class Font
	{
		private Point position = new Point(0, 0);
		private Size size;
		private Bitmap[] characters = new Bitmap[59];

		public Font(Bitmap bitmap, Size size)
		{
			this.size = size;
			Rectangle cursor = new Rectangle(0, 0, size.Width, size.Height);
			for (int i = 0; i < 59; i++)
			{
				this.characters[i] = bitmap.Clone(cursor, PixelFormat.Format24bppRgb);
				this.characters[i].MakeTransparent(Color.FromArgb(0, 0, 0));
				cursor.Y += cursor.Height;
			}
		}

		public void Reset()
		{
			this.position = new Point(0, 0);
		}

		public void Write(Graphics graphics, string text)
		{
			SolidBrush redBrush = new SolidBrush(Color.FromArgb(0x92, 0x02, 0x05));
			for (int i = 0; i < text.Length; i++) 
			{
				graphics.FillRectangle(redBrush, new Rectangle(this.position.X, this.position.Y, this.size.Width, this.size.Height));
				graphics.DrawImageUnscaled(this.characters[(byte) (text[i] - 32)], this.position.X, this.position.Y);
				this.position.X += size.Width;
			}
		}

		public void Newline()
		{
			this.position.X = 0;
			this.position.Y += size.Height;
		}
	}
}
